package com.almworks.jira.structure.customitemtype;

import com.almworks.jira.structure.api.item.ItemIdentity;
import com.almworks.jira.structure.api.item.ItemTracker;
import com.almworks.jira.structure.api.item.StructureItemType;
import com.atlassian.event.api.EventListener;
import com.atlassian.event.api.EventPublisher;
import com.atlassian.jira.event.AbstractProjectEvent;
import com.atlassian.jira.permission.ProjectPermissions;
import com.atlassian.jira.project.Project;
import com.atlassian.jira.project.ProjectManager;
import com.atlassian.jira.security.PermissionManager;
import com.atlassian.jira.user.ApplicationUser;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;

/**
 * <p>Implementation of CustomProject item type. This class demonstrates how to implement basic Structure item type:</p>
 * <ul>
 *   <li>Expose an entity (which is stored somehow in JIRA) into a Structure</li>
 *   <li>{@link #onProjectEvent(AbstractProjectEvent) Notifies} the Structure plugin when a project has been changed</li>
 *   <li>{@link #isVisible(Project, ApplicationUser) Limits} access to CustomProject items in the Structure for some users</li>
 * </ul>
 */
public class CustomProjectItemType implements StructureItemType<Project>, InitializingBean, DisposableBean {
  /**
   * Value of the 'key' attribute of the 'atlassian-plugin' tag in the 'atlassian-plugin.xml' file.
   */
  public static final String PLUGIN_KEY = "com.almworks.jira.structure.customitemtype";
  /**
   * Value of the 'key' attribute of the 'structure-item-type' module defined in the 'atlassian-plugin.xml' file.
   */
  public static final String COMPONENT_KEY = "custom-project-itemtype";
  /**
   * This item type string identifier.
   * Structure automatically registers a new item type and constructs its string identifier based on value from the 'atlassian-plugin.xml' file.
   */
  public static final String KEY = PLUGIN_KEY + ":" + COMPONENT_KEY;

  private final ProjectManager myProjectManager;
  private final PermissionManager myPermissionManager;
  private final ItemTracker myItemTracker;
  private final EventPublisher myEventPublisher;

  /**
   * @param itemTracker don't forget to add 'component-import' to the 'atlassian-plugin.xml' file
   */
  public CustomProjectItemType(ProjectManager projectManager, PermissionManager permissionManager, ItemTracker itemTracker, EventPublisher eventPublisher) {
    myProjectManager = projectManager;
    myPermissionManager = permissionManager;
    myItemTracker = itemTracker;
    myEventPublisher = eventPublisher;
  }

  @Override
  public void destroy() throws Exception {
    myEventPublisher.unregister(this);
  }

  @Override
  public void afterPropertiesSet() throws Exception {
    myEventPublisher.register(this);
  }

  /**
   * <p>Listens to JIRA project updated events and notifies the Structure plugin.
   * This method makes the Structure plugin reload attribute values of a CustomProject item that corresponds to the updated JIRA project.</p>
   * <p>Also this method handles delete and create events. They are essential to detect project set change with {@link com.almworks.jira.structure.api.generator.ItemChangeFilter}</p>
   * @see CustomProjectInserter#createItemChangeFilter()
   */
  @EventListener
  public void onProjectEvent(AbstractProjectEvent event) {
    myItemTracker.recordChange(createId(event.getProject()));
  }

  /**
   * A convenient method to convert JIRA project to a Structure CustomProject item {@link ItemIdentity identity}.
   * @param project a JIRA project
   * @return Structure item identity
   */
  public static ItemIdentity createId(Project project) {
    return ItemIdentity.longId(KEY, project.getId());
  }

  @Nullable
  public Project accessItem(@NotNull ItemIdentity itemId) {
    if (!KEY.equals(itemId.getItemType())) return null;
    if (itemId.isLongId()) return myProjectManager.getProjectObj(itemId.getLongId());
    else return null;
  }

  /**
   * Hides projects from a user who has no permission to browse the projects.
   * @param project a JIRA project to check user's permissions for
   * @param user a user to check permissions
   * @return true if the user is permitted to browse the project
   */
  public boolean isVisible(@NotNull Project project, @Nullable ApplicationUser user) {
    return myPermissionManager.hasPermission(ProjectPermissions.BROWSE_PROJECTS, project, user);
  }
}
