package com.almworks.jira.structure.scheduledsync;

import com.almworks.integers.LongArray;
import com.almworks.integers.LongList;
import com.almworks.integers.segmented.LongSegmentedArray;
import com.almworks.jira.structure.api.process.ProcessHandle;
import com.atlassian.activeobjects.external.ActiveObjects;
import com.atlassian.cache.*;
import com.atlassian.fugue.Option;
import net.java.ao.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.sql.SQLException;

/**
 * @author Igor Sereda
 * @author Igor Baltiyskiy
 */
public final class Schedule {
  private static final Logger logger = LoggerFactory.getLogger(Schedule.class);

  public static final String PLUGIN_KEY = "com.almworks.jira.structure.scheduledsync";
  public static final String PROPERTY_PREFIX = PLUGIN_KEY + ".";

  private final ActiveObjects myActiveObjects;
  private final Cache<Long, ScheduledSync> myData;
  private final ScheduledSyncIO myScheduledSyncIO;
  private final CachedReference<LongList> myAllSyncs;

  public Schedule(ActiveObjects activeObjects, CacheManager cacheManager) {
    myActiveObjects = activeObjects;
    myScheduledSyncIO = new ScheduledSyncIO();
    myData = cacheManager.getCache(PROPERTY_PREFIX + "data", myScheduledSyncIO);
    myAllSyncs = cacheManager.getCachedReference(PROPERTY_PREFIX + "allSyncs", new AllSyncsLoader());
  }
  
  public String toString() {
    return "SyncSchedule";
  }

  /** 
   * Returns true if schedule was created or deleted. 
   * False otherwise, including: if only modified an existing schedule, or if attempted to delete a non-existing one. */
  public boolean setSchedule(long syncId, long period) {
    if (period <= 0) {
      if (myScheduledSyncIO.remove(syncId)) {
        logger.info(this + " removed " + syncId + " from schedule");
        return true;
      }
    } else if (period >= 10000) {
      boolean created = myScheduledSyncIO.createOrChange(syncId, period);
      logger.info(this + " scheduled " + syncId + " to run each " + period + " ms");
      return created;
    } else {
      logger.warn("period too small: {}, not changing schedule for {}", period, syncId);
    }
    return false;
  }

  public Option<ScheduledSync> getSchedule(long syncId) {
    try {
      return Option.option(myData.get(syncId));
    } catch (ActiveObjectsException|CacheException e) {
      if (e.getCause() instanceof NoSyncException) {
        // ignore
      } else {
        logger.info("No schedule for #" + syncId);
      }
      return Option.none();
    }
  }

  public void setResyncProcess(long syncId, ProcessHandle process) {
    myScheduledSyncIO.setProcessId(syncId, process.getId());
  }

  public LongList getAllScheduledSyncs() {
    return myAllSyncs.get();
  }

  public static class ScheduledSync {
    private final long mySyncId;
    private final long myPeriod;
    private final long myLastResyncProcessId;

    public ScheduledSync(long syncId, long period, long lastResyncProcessId) {
      if (syncId <= 0) throw new IllegalArgumentException("syncId " + syncId);
      if (period < 10000) throw new IllegalArgumentException("period " + period);
      mySyncId = syncId;
      myPeriod = period;
      myLastResyncProcessId = lastResyncProcessId;
    }

    public String toString() {
      return "scheduled-sync(" + mySyncId + ", period: " + myPeriod + "ms)";
    }

    public long getSyncId() {
      return mySyncId;
    }

    public long getPeriod() {
      return myPeriod;
    }

    public long getLastResyncProcessId() {
      return myLastResyncProcessId;
    }
  }
  
  
  private class ScheduledSyncIO implements CacheLoader<Long, ScheduledSync> {
    @Nonnull
    @Override
    public ScheduledSync load(@Nonnull Long syncId) {
      ScheduledSyncAO ao = myActiveObjects.get(ScheduledSyncAO.class, syncId);
      if (ao == null) throw new NoSyncException();
      return new ScheduledSync(syncId, ao.getPeriod(), ao.getLastResyncProcessId());
    }

    public boolean createOrChange(long syncId, long period) {      
      ScheduledSyncAO ao = myActiveObjects.get(ScheduledSyncAO.class, syncId);
      boolean created = true;
      try {
        if (ao == null) {
          myActiveObjects.create(ScheduledSyncAO.class,
            new DBParam(ScheduledSyncAO.SYNC_ID, syncId),
            new DBParam(ScheduledSyncAO.PERIOD, period),
            new DBParam(ScheduledSyncAO.LAST_RESYNC_PROCESS_ID, 0L));
          created = true;
        } else {
          ao.setPeriod(period);
          ao.save();
          created = false;
        }
      } catch (ActiveObjectsException e) {
        if (isConstraintViolation(e)) {
          // ignore
          logger.info("Sync schedule for {} already exists", syncId);
        } else {
          throw e;
        }        
      } finally {
        myData.remove(syncId);
        if (created) {
          myAllSyncs.reset();
        }
      }
      return created;
    }

    public void setProcessId(long syncId, long processId) {
      ScheduledSyncAO ao = myActiveObjects.get(ScheduledSyncAO.class, syncId);
      try {
        ao.setLastResyncProcessId(processId);
        ao.save();
      } finally {
        myData.remove(syncId);
      }
    } 

    private boolean isConstraintViolation(ActiveObjectsException e) {
      SQLException sqlex = e.getCause() instanceof SQLException ? (SQLException) e.getCause() : null;
      while (sqlex != null) {
        if (sqlex.getErrorCode() == 20000) return true;
        String sqlState = sqlex.getSQLState();
        // See SQL:2003 error codes
        if (sqlState != null && sqlState.startsWith("23")) {
          return true;
        }
        sqlex = sqlex.getNextException();
      }
      return false;
    }

    public boolean remove(long syncId) {
      ScheduledSyncAO ao = myActiveObjects.get(ScheduledSyncAO.class, syncId);
      if (ao != null) {
        try {
          myActiveObjects.delete(ao);
        } finally {
          myData.remove(syncId);
          myAllSyncs.reset();
        }
        return true;
      }
      return false;
    }
  }
  
  private class AllSyncsLoader implements Supplier<LongList> {
    @Override
    public LongList get() {
      final LongSegmentedArray syncIds = new LongSegmentedArray();
      myActiveObjects.stream(ScheduledSyncAO.class, new EntityStreamCallback<ScheduledSyncAO, Long>() {
        @Override
        public void onRowRead(ScheduledSyncAO ao) {
          syncIds.add(ao.getSyncId());
        }
      });
      return new LongArray(syncIds);
    }
  }


  private class NoSyncException extends RuntimeException {
  }  
}
