package com.almworks.jira.structure.scheduledsync;

import com.almworks.jira.structure.api.StructureComponents;
import com.almworks.jira.structure.api.permissions.PermissionLevel;
import com.almworks.jira.structure.api.structure.Structure;
import com.almworks.jira.structure.api.structure.StructureManager;
import com.almworks.jira.structure.api.sync.StructureSyncManager;
import com.almworks.jira.structure.api.sync.SyncInstance;
import com.almworks.jira.structure.api.util.StructureUtil;
import com.atlassian.jira.security.Permissions;
import com.atlassian.jira.security.xsrf.RequiresXsrfCheck;
import com.atlassian.jira.web.action.JiraWebActionSupport;
import org.jetbrains.annotations.Nullable;
import webwork.action.ActionContext;
import webwork.action.ResultException;

import java.text.DateFormat;
import java.util.*;

/**
 * @author Igor Sereda
 */
public class StructureScheduledResync extends JiraWebActionSupport {
  private final Scheduler myScheduler;
  private final StructureSyncManager mySyncManager;
  private final StructureManager myStructureManager;
  private long myNow;

  public StructureScheduledResync(Scheduler scheduler, StructureComponents services) {
    myScheduler = scheduler;
    mySyncManager = services.getSyncManager();
    myStructureManager = services.getStructureManager();
  }

  protected String doExecute() throws Exception {
    if (!hasPermission(Permissions.ADMINISTER)) return "securitybreach";
    myNow = System.currentTimeMillis();
    return SUCCESS;
  }

  @RequiresXsrfCheck
  public final String doApply() throws ResultException {
    if (!hasPermission(Permissions.ADMINISTER)) return "securitybreach";
    apply();
    return getRedirect("/secure/StructureScheduledResync.jspa");
  }

  private void apply() {
    List<Structure> structures = myStructureManager.getAllStructures(PermissionLevel.ADMIN);
    Map parameters = ActionContext.getParameters();
    for (Structure structure : structures) {
      List<SyncInstance> syncs = mySyncManager.getInstalledSynchronizersForStructure(structure.getId());
      for (SyncInstance sync : syncs) {
        Schedule.ScheduledSync schedule = myScheduler.getSchedule(sync.getId());
        Long period = StructureUtil.getSingleParameterLong(parameters, "period_" + sync.getId());
        if (period != null) {
          if (period > 0) {
            long ms = period * 60000;
            if (schedule == null || schedule.getPeriod() != ms) {
              myScheduler.setSchedule(sync.getId(), ms);
            }
          } else {
            if (schedule != null) {
              myScheduler.setSchedule(sync.getId(), 0);
            }
          }
        }
      }
    }
  }

  public List<SyncInfo> getSyncs() {
    List<SyncInfo> r = new ArrayList<SyncInfo>();
    List<Structure> structures = myStructureManager.getAllStructures(PermissionLevel.ADMIN);
    for (Structure structure : structures) {
      List<SyncInstance> syncs = mySyncManager.getInstalledSynchronizersForStructure(structure.getId());
      for (SyncInstance sync : syncs) {
        if (sync.getSynchronizer() == null) continue;
        Schedule.ScheduledSync schedule = myScheduler.getSchedule(sync.getId());
        r.add(new SyncInfo(structure, sync, schedule, mySyncManager.isAutosyncEnabled(sync.getId())));
      }
    }
    return r;
  }


  public class SyncInfo {
    private final Structure myStructure;
    private final SyncInstance mySync;
    @Nullable
    private final Schedule.ScheduledSync mySchedule;
    private final long myLastRun;
    private final long myNextRun;
    private final boolean myAutosyncEnabled;

    public SyncInfo(Structure structure, SyncInstance sync, @Nullable Schedule.ScheduledSync schedule, boolean autosyncEnabled) {
      myStructure = structure;
      mySync = sync;
      mySchedule = schedule;
      myLastRun = myScheduler.getLastRunTime(schedule);
      myNextRun = myScheduler.getNextRunTime(schedule, myLastRun, myNow);
      myAutosyncEnabled = autosyncEnabled;
    }

    public int getMinutesPeriod() {
      return mySchedule == null ? 0 : (int)(mySchedule.getPeriod() / 60000);
    }

    public boolean isAutosyncEnabled() {
      return myAutosyncEnabled;
    }

    public Structure getStructure() {
      return myStructure;
    }

    public SyncInstance getSync() {
      return mySync;
    }
    
    public long getLastResyncProcessId() {
      return mySchedule == null ? 0 : mySchedule.getLastResyncProcessId();
    }

    public boolean hasLastRun() {
      return getLastResyncProcessId() > 0;
    }

    public String getLastRunDate() {
      return hasLastRun() ? dateFormat().format(new Date(myLastRun)) : "";
    }    

    private DateFormat dateFormat() {
      return DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT);
    }

    public boolean hasNextRun() {
      return mySchedule != null && myNextRun > 0;
    }

    public String getNextRunDate() {
      return hasNextRun() ? dateFormat().format(new Date(myNextRun)) : "";
    }
  }
}
