package com.almworks.jira.structure.sbcolumn;

import com.almworks.jira.structure.api.attribute.CoreAttributeSpecs;
import com.almworks.jira.structure.api.error.StructureProviderException;
import com.almworks.jira.structure.api.export.*;
import com.almworks.jira.structure.api.export.excel.ExcelCell;
import com.almworks.jira.structure.api.export.excel.ExcelColumn;
import com.almworks.jira.structure.api.util.StructureUtil;
import com.almworks.jira.structure.api.view.ViewSpecification;
import com.atlassian.jira.issue.status.Status;
import org.apache.commons.lang.StringUtils;
import org.apache.poi.hssf.usermodel.*;
import org.apache.poi.hssf.util.HSSFColor;
import org.apache.poi.ss.usermodel.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;

/**
 * Excel renderer provider for the Status Bar column.
 * This class shows how you can bypass the {@link ExcelCell} and {@link ExcelColumn} API and use Apache POI
 * directly to do advanced formatting in Excel exports.
 */
public class StatusBarExcelProvider implements ExportRendererProvider<ExcelCell, ExcelColumn> {
  private static final int BAR_LENGTH = 30;
  private static final String BAR_STRING = StringUtils.repeat("|", BAR_LENGTH);

  @Nullable
  @Override
  public ExportRenderer<ExcelCell, ExcelColumn> getColumnRenderer(
    @NotNull ExportFormat exportFormat, @NotNull ViewSpecification.Column columnSpec, @NotNull ExportRequestContext context)
    throws StructureProviderException
  {
    if (exportFormat != ExportFormat.MS_EXCEL || !StatusBarRendererProvider.COLUMN_KEY.equals(columnSpec.getKey())) {
      return null;
    }

    String name = columnSpec.getName();
    if (StringUtils.isEmpty(name)) {
      name = context.getI18n().getText("sbcolumn.name");
    }

    Map<String, Object> parameters = columnSpec.getParameters();

    List<String> statuses = StructureUtil.getMultiParameter(parameters, "statuses");
    if (statuses.isEmpty()) {
      throw new StructureProviderException("statuses parameter is required");
    }
    List<String> colors = StructureUtil.getMultiParameter(parameters, "colors");
    if (colors.isEmpty()) {
      throw new StructureProviderException("colors parameter is required");
    }
    if (statuses.size() != colors.size()) {
      throw new StructureProviderException("colors.length must match statuses.length");
    }

    boolean includeItself = StructureUtil.getSingleParameterBoolean(parameters, "includeItself");

    return new StatusBarRenderer(name, statuses, colors, includeItself);
  }


  private static class StatusBarRenderer implements ExportRenderer<ExcelCell, ExcelColumn> {
    private final String myName;
    private final List<String> myStatuses;
    private final List<String> myColors;

    private final Map<String, Integer> myExclude;
    private final Map<String, Integer> myTempMap = new HashMap<String, Integer>();

    public StatusBarRenderer(String name, List<String> statuses, List<String> colors, boolean includeItself) {
      myName = name;
      myStatuses = statuses;
      myColors = colors;
      myExclude = includeItself ? null : new HashMap<String, Integer>();
    }

    @Override
    public void prepare(@NotNull ExportRequestContext context) {
      context.requireAttribute(StatusBarAttributeProvider.STATUS_BAR);
      context.requireAttribute(CoreAttributeSpecs.STATUS);
    }

    @Override
    public void configureColumn(@NotNull ExcelColumn column, @NotNull ExportRenderContext context) {
      column.getHeaderCell().setText(myName);
    }

    @Override
    public void renderCell(@NotNull ExcelCell excelCell, @NotNull ExportRow row, @NotNull ExportRenderContext context) {
      Map<String, Integer> values = row.get(StatusBarAttributeProvider.STATUS_BAR);
      if (values == null || values.isEmpty()) {
        return;
      }

      if (myExclude != null) {
        if (row.isLeaf()) {
          return;
        }
        values = excludeItself(values, row.get(CoreAttributeSpecs.STATUS));
      }
      values = distribute(values, BAR_LENGTH);
      renderStatusBar(values, context);
    }

    private Map<String, Integer> excludeItself(Map<String, Integer> values, Status selfStatus) {
      if (selfStatus == null) {
        return values;
      }
      myExclude.clear();
      myExclude.putAll(values);
      String status = selfStatus.getId();
      Integer value = myExclude.get(status);
      if (value != null && value > 0) {
        myExclude.put(status, value - 1);
      }
      return myExclude;
    }

    private Map<String, Integer> distribute(Map<String, Integer> values, int amount) {
      float total = 0;
      int max = 0;
      for (String key : myStatuses) {
        Integer value = values.get(key);
        if (value != null && value > 0) {
          total += value;
          max = Math.max(max, value);
        }
      }

      if (total == 0) {
        return Collections.emptyMap();
      }

      Map<String, Integer> result = myTempMap;
      result.clear();
      int remaining = amount;
      for (String key : myStatuses) {
        Integer value = values.get(key);
        if (value != null && value > 0 && value < max) {
          int p = Math.round(value.floatValue() / total * amount);
          result.put(key, p);
          remaining -= p;
        }
      }

      String last = null;
      for (String key : myStatuses) {
        Integer value = values.get(key);
        if (value != null && value == max) {
          int p = Math.min(remaining, Math.round(value.floatValue() / total * amount));
          result.put(key, p);
          remaining -= p;
          last = key;
        }
      }

      if (remaining > 0) {
        result.put(last, result.get(last) + remaining);
      }

      return result;
    }

    private void renderStatusBar(Map<String, Integer> values, ExportRenderContext context) {
      Palette palette = context.getObject(Palette.class);
      if (palette == null) {
        HSSFWorkbook workbook = context.getObject(ExportContextKeys.Excel.POI_WORKBOOK);
        palette = new Palette(workbook);
        context.putObject(Palette.class, palette);
      }

      Cell cell = context.getObject(ExportContextKeys.Excel.POI_CELL);
      CreationHelper helper = context.getObject(ExportContextKeys.Excel.POI_HELPER);
      RichTextString bar = helper.createRichTextString(BAR_STRING);

      int start = 0;
      for (int i = 0; i < myStatuses.size(); i++) {
        Integer value = values.get(myStatuses.get(i));
        if (value != null && value > 0) {
          if (start == 0) {
            palette.setCellStyle(cell, myColors.get(i));
          } else {
            palette.applyFont(bar, start, start + value, myColors.get(i));
          }
          start += value;
        }
      }

      cell.setCellValue(bar);
    }
  }


  private static class Palette {
    private final HSSFWorkbook myWorkbook;
    private final HSSFPalette myPalette;

    private final Map<String, Short> myColors = new HashMap<String, Short>();
    private final Map<String, Short> myFonts = new HashMap<String, Short>();
    private final Map<String, CellStyle> myStyles = new HashMap<String, CellStyle>();

    public Palette(HSSFWorkbook workbook) {
      myWorkbook = workbook;
      myPalette = workbook.getCustomPalette();
    }

    private short getColorIndex(String hex) {
      Short index = myColors.get(hex);
      if (index == null) {
        RGB rgb = RGB.fromHex(hex);
        HSSFColor color = getHSSFColor(rgb);
        index = color.getIndex();
        myColors.put(hex, index);
      }
      return index;
    }

    private HSSFColor getHSSFColor(RGB rgb) {
      HSSFColor color = myPalette.findColor(rgb.red, rgb.green, rgb.blue);
      if (color == null) {
        color = myPalette.findSimilarColor(rgb.red, rgb.green, rgb.blue);
        if (color == null) {
          try {
            color = myPalette.addColor(rgb.red, rgb.green, rgb.blue);
          } catch (RuntimeException ignored) {
            // palette full
          }
        }
      }
      return color != null ? color : getHSSFColor(RGB.BLACK);
    }

    private short getFontIndex(String hex) {
      Short index = myFonts.get(hex);
      if (index == null) {
        HSSFFont font = myWorkbook.createFont();
        font.setColor(getColorIndex(hex));
        index = font.getIndex();
        myFonts.put(hex, index);
      }
      return index;
    }

    private HSSFFont getFont(String hex) {
      return myWorkbook.getFontAt(getFontIndex(hex));
    }

    public void setCellStyle(Cell cell, String hex) {
      CellStyle style = myStyles.get(hex);
      if (style == null) {
        style = myWorkbook.createCellStyle();
        style.cloneStyleFrom(cell.getCellStyle());
        style.setFont(getFont(hex));
        myStyles.put(hex, style);
      }
      cell.setCellStyle(style);
    }

    public void applyFont(RichTextString string, int from, int to, String hex) {
      string.applyFont(from, to, getFontIndex(hex));
    }
  }


  private static class RGB {
    static final RGB BLACK = new RGB(0, 0, 0);

    final byte red;
    final byte green;
    final byte blue;

    public RGB(int red, int green, int blue) {
      this.red = (byte) red;
      this.green = (byte) green;
      this.blue = (byte) blue;
    }

    static RGB fromHex(String hex) {
      try {
        int rgb = Integer.decode(hex);
        return new RGB((rgb >> 16) & 0xFF, (rgb >> 8) & 0xFF, rgb & 0xFF);
      } catch (NumberFormatException e) {
        return BLACK;
      }
    }
  }
}
